#define _CRT_SECURE_NO_WARNINGS

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "xmalloc.h"
#include "object.h"
#include "util.h"
#include "Attribute.h"
#include "AttributeCache.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "EntityInstance.h"
#include "boundaries.h"

static int	g_max_args = -1;
static int	g_args     = 0;
char		**g_argsPP = 0;
static char	 *g_argsP  = 0;

static int	g_boundaries_max = 0;
int			g_boundaries     = 0;
boundaryT	*g_boundariesP   = 0;

static void trap2()
{
}

static void
recover_args(const char *cargsP)
{
	char	*P, *P1, c;
	char 	**argsPP, **args1PP, *argP;
	int		load, state;

	Xfree(g_argsP);
	g_argsP  = dupstring(cargsP);

	for (load = 0; ; ++load) {
		g_args = 0;
		state  = 0;
		for (P = g_argsP; ; ++P) {
			switch ((c = *P)) {
			case 0:
				break;
			case '"':
				switch (state) {
				case 0:
					if (load) {
						*argsPP = P1 = P+1;
					}
					state = 1;
					break;
				case 1:
					goto loaded;
				default:
					assert(0);
				}
				continue;
			case '\n':	// The following are all quoted by lsedit
			case '\t':	// each therefore terminates an unquoted value
			case '\f':
			case '\r':
			case '\x1b':
			case '\x7f':
			case ' ':
			case ',':
			case '(':
			case ')':
			case '{':
			case '}':
				if (state == 2) {
loaded: 			if (load) {
						*P1 = 0;
						++argsPP;
					}
					state = 0;
					if (++g_args == 2) {
						if (load) {
							*argsPP++ = "-E";
							*argsPP++ = "-ftabstop=1"; /* Dont expand tabs */
						}
						g_args += 2;
				}	}
				continue;
			default:
				if (!state) {
					if (load) {
						*argsPP = P1 = P;
					}
					state = 2;
				}
				if (c == '\\') {
					++P;
					switch (c = *P) {
					case 'r':
						c = '\r';
						break;
					case 'n':
						c = '\n';
						break;
					case 0:
						assert(0);
				}	}
				if (load) {
					*P1++ = c;
				}
				continue;
			}
			break;
		}

		if (load) {
			break;
		}
		assert (g_args >= 3);

		if (g_args > g_max_args) {
			Xfree(g_argsPP);
			g_argsPP   = (char **) Xmalloc(sizeof(char *) * (g_args+2));
			g_max_args = g_args;
		}
		argsPP = g_argsPP;
	}

	assert(!state);
	*argsPP = 0;

	for (argsPP = args1PP = g_argsPP+2; (argP = *argsPP); ++argsPP) {
		if (*argP == '-') {
			switch (argP[1]) {
			case 'o':
				if (!argP[2]) {
					++argsPP;
				}
			case 'P':
			case 'C':
			case 'S':
			case 'g':
			case 'c':
				continue;
		}	}
		*args1PP++ = argP;
	}
	*args1PP = 0;
}

/* Scan the postprocessed source */

static void
scan(const char *cwdP, int fd)
{
	extern char	*g_filenameP;
	extern int	g_debug_flag;

	FILE		*F;
	int			c, lastc;
	int			lineno, character;
	int			quote, val;
	int			state, function;
	char		filename[2048];
	char		*filenameP;
	boundaryT	*boundaryP;
	FILE		*outF;
	int			openpoint, functionptr;
	int			cwd_lth;

	outF    = 0;
	cwd_lth = -1;
	if (g_debug_flag) {
		sprintf(filename, "%s.scan", g_filenameP);
		outF = fopen(filename, "w");
		if (!outF) {
			fprintf(stderr, "Can't create %s\n", filename);
		} else {
			fprintf(outF, "0001 ");
	}	}

	F = fdopen(fd, "r");
	if (!F) {
		fprintf(stderr, "Can't open input pipe from cpp\n");
		error_exit(2);
	}

	if (!g_boundariesP) {
		g_boundaries_max = 2048;
		g_boundariesP    = (boundaryT *) Xmalloc(sizeof(boundaryT) * g_boundaries_max);
	}

	boundaryP    = 0;
	lineno       = 1;
	character    = 0;
	state        = 0;
	quote        = 0;
	function     = -1;
	functionptr  = 0;
	g_boundaries = 0;
	openpoint    = 0;
	lastc        = 0;

	while ((c = getc(F)) != EOF) {
		++character;

		// Track status of within quotes
		switch (c) {
		case '"':
			switch (quote) {
			case 0:
				quote = 2;
				break;
			case 2:
				quote = 0;
				break;
			}
			break;
		case '\'':
			switch (quote) {
			case 0:
				quote = 1;
				break;
			case 1:
				quote = 0;
				break;
			}
			break;
		case '\\':
			c = getc(F);
			if (c == EOF) {
				goto done;
			}
			break;
		case '\n':
			character = 0;
		}
	

		// Track # lineno statements

		switch (state) {
		case 0:					// At start of line
			if (!quote && c == '#') {
				state = 1;
				goto advance;
			}
			state = -1;
			break;
		case 1:
			if (c == ' ') {
				goto advance;
			}
			if (c < '1' || c > '9') {
				state = -1;
				break;
			}
			val   = c - '0';
			state = 2;
			goto advance;
		case 2:
			if (c >= '0' && c <= '9') {
				val = val * 10 + c - '0';
				goto advance;
			}
			if (c != ' ') {
				state = -1;
				break;
			}
			state = 3;
			goto advance;
		case 3:
			if (c == ' ') {
				goto advance;
			}
			if (c != '"') {
				state = -1;
				break;
			}
			state = 4;
			filenameP = filename;
			goto advance;
		case 4:
			if (c != '"') {
				*filenameP++ = c;
				goto advance;
			}
			*filenameP = 0;
			state = 5;
			goto advance;
		case 5:
		case 6:
/* Following flags:
	1 - This indicates the start of a new file
    2 - This indicates return to a file
    3 - This indicates that the following text is from a system header
    4 - Following text treat as wrapped in extern "C" block
 */
			switch (c) {
			case '\n':
				state     = 0;
				switch(filename[0]) {
				case '<':
				case '/':
					break;
				default:
					if (cwd_lth < 0) {
						cwd_lth = strlen(cwdP);
					}
					memmove(filename+cwd_lth+1, filename, filenameP-filename+1);
					memcpy(filename, cwdP, cwd_lth);
					filename[cwd_lth] = '/';
				}
				if (!strcmp(filename, g_filenameP)) {
					lineno = val;
					if (outF) {
						fprintf(outF, "\n%04d ", lineno);
					}
				} else {
					lineno = -1;
					if (outF) {
						fprintf(outF, "\n     ");
				}	}
				continue;
			case '\r':
			case '\t':
			case ' ':
				goto advance;
			default:
				state = 6;
				goto advance;
		}	}

		// Track normal C/C++ code
		if (quote) {
			goto advance;
		}

		if (lastc == '(' && !function && !functionptr)  {
			if (c != '*') {
				openpoint = g_boundaries;
			} else {
				functionptr = 1;
		}	}

		switch (c) {
		case '\n':
			if (lineno > 0) {
				++lineno;
				if (outF) {
					fprintf(outF, "\n%04d+", lineno);
				}
			} else {
				if (outF) {
					fprintf(outF, "\n     ");
			}	}
			state     = 0;
			continue;
		case ' ':
		case '\t':
		case '\r':
			// Don't update lastc
			goto advance;
		case '(':
			if (functionptr) {
				++functionptr;
			}
			goto updatelast;
		case ')':
			if (functionptr) {
				--functionptr;
			}
			goto updatelast;
		case '{':
			if (0 <= function) {
				if (!openpoint) {
					// Discard this start -- perhaps class .. { or struct .. {
					--g_boundaries;
					function  = -1;
					boundaryP = 0;
				} else {
					++function;			// Count brackets
				}
			} else if (openpoint && lastc == ';') {
				// Think we are in the strange world of ancient C code
				// where one could write:
				// int function(x, y)
				//	int x;
				//	int y;
				// {

				trap2();
				if (outF) {
					fprintf(outF, "!!");
				}

				// Compact boundaries back to openpoint
				boundaryP    = g_boundariesP + openpoint - 1;
				g_boundaries = openpoint;
				function     = 1;
			}
			goto updatelast;
		case '}':
			if (function > 0) {
				if (!--function) {
					openpoint = 0;
					goto end_boundary;
			}	}
			goto updatelast;
		case ';':
			if (!function ) {
end_boundary:
				if (outF) {
					fputc(c, outF);
				}

				if (lineno > 0 && boundaryP) {
					if (outF) {
						fprintf(outF, "\n##END lineno=%d+%d\n", lineno, character);
					}
					boundaryP->lineno[1]    = lineno;
					boundaryP->character[1] = character+1;
					boundaryP               = 0;
				}
				function  = -1;
				lastc     = c;
				continue;
			}
			goto updatelast;
		}

		if (function == -1) {
			// start of function
			if (lineno > 0 && !boundaryP) {
				function  = 0;
				if (outF) {
					fprintf(outF, "\n##START lineno=%d+%d\n", lineno, character);
				}
				if (g_boundaries >= g_boundaries_max) {
					g_boundaries_max <<= 1;
					g_boundariesP      = (boundaryT *) Xrealloc(g_boundariesP, sizeof(boundaryT) * g_boundaries_max);
				}
				boundaryP = g_boundariesP + g_boundaries;
				boundaryP->lineno[0]    = lineno;
				boundaryP->character[0] = character;
				boundaryP->lineno[1]    = -1;
				boundaryP->dead         = 0;
				++g_boundaries;
		}	}
updatelast:
		lastc = c;
advance:
		if (outF) {
			fputc(c, outF);
		}
		continue;
	}
done:
	fclose(F);
	if (outF) {
		fprintf(outF, "\n\n");
		for (boundaryP = g_boundariesP; boundaryP < g_boundariesP + g_boundaries; ++boundaryP) {
			fprintf(outF, "%d.%d - %d.%d\n", boundaryP->lineno[0], boundaryP->character[0], boundaryP->lineno[1], boundaryP->character[1]);
		}
		fclose(outF);
	}
}

static int
call_cc(char **argsPP)
{
	extern int g_trace_flag;

	int		fds[2];

	char	*argP;
	char	*cwdP;
	char	*programP;
	int		ret, pid;
	int		status;
	int 	i;

	if (g_trace_flag) {
		for (i = 0; (argP = argsPP[i]); ++i) {
			fprintf(stderr, "%s ", argP);
		}
		fprintf(stderr,"\n");
	}
	
	cwdP     = *argsPP++;
	programP = *argsPP;

	ret = pipe(fds);
	if (ret) {
		fprintf(stderr, "Can't create pipe ret=%d\n", ret);
		error_exit(1);
	}
	ret = fork();
	switch (ret) {
	case 0:
		if (chdir(cwdP)) {
			fprintf(stderr,"Can't change directory to %s\n", cwdP);
			perror("");
			return(-1);
		}
		// Make standard output write to pipe
		close(fds[0]);
		ret = dup2(fds[1], 1);
		if (ret == -1) {
			fprintf(stderr, "Write to pipe (dup2) failed ret=%d\n", ret);
			perror("");
			error_exit(1);
		}
		execve(programP, argsPP, 0);
		fprintf(stderr,"Return from execve %s\n", programP);
		error_exit(1);
	case -1:
		fprintf(stderr,"Fork to run %s failed\n", programP);
		perror("");
		error_exit(2);
	}

	close(fds[1]);
	scan(cwdP, fds[0]);

	pid = wait(&status);
	if (pid != ret) {
		fprintf(stderr,"Wait for %s failed\n", programP);
		perror("");
		error_exit(3);
	}
	if (!WIFEXITED(status)) {
		fprintf(stderr,"%s did not terminate normally\n", programP);
		return(-1);
	}
	return(WEXITSTATUS(status));
}

int
establish_boundaries(CEntityInstance *eP)
{
	CEntityInstance	*sourceP;
	const char		*argsP;

	sourceP = eP->getParent();
	if (!sourceP) {
		return(0) ;
	}
	argsP   = sourceP->getAttribute("args");
	if (!argsP) {
		return(0);
	}
	recover_args(argsP);
	call_cc(g_argsPP);
	return (g_boundaries);
}
